from os.path import basename

import psycopg2
from django.contrib.gis.geos import GEOSGeometry
from django.db import transaction
from django.utils.translation import ugettext as _
from psycopg2 import sql
from rest_framework.serializers import (
    BooleanField,
    CharField,
    ChoiceField,
    IntegerField,
    ModelSerializer,
    SerializerMethodField,
    SlugField,
    ValidationError,
)

from .models import (
    CommandSource,
    CSVSource,
    Field,
    GeoJSONSource,
    GeometryTypes,
    PostGISSource,
    ShapefileSource,
    Source,
    WMTSSource,
)


class PolymorphicModelSerializer(ModelSerializer):
    type_field = "_type"
    type_class_map = {}

    def __new__(cls, *args, **kwargs):
        """
        Return the correct serializer given the type provided in type_field
        """
        if kwargs.pop("many", False):
            return cls.many_init(*args, **kwargs)

        if "data" in kwargs:

            data_type = kwargs["data"].get(cls.type_field)

            serializer = cls.get_serializer_from_type(data_type)

            if serializer is not cls:
                return serializer(*args, **kwargs)

        return super().__new__(cls, *args, **kwargs)

    def __init_subclass__(cls, **kwargs):
        """ Create a registry of all subclasses of the current class """
        if cls.Meta.model:
            cls.type_class_map[cls.Meta.model.__name__] = cls

    @classmethod
    def get_serializer_from_type(cls, data_type):
        """
        Returns the serializer class from datatype
        """
        if data_type in cls.type_class_map:
            return cls.type_class_map[data_type]
        raise ValidationError({cls.type_field: f"{data_type}'s type is unknown"})

    def to_representation(self, obj):
        serializer = self.get_serializer_from_type(obj.__class__.__name__)

        if serializer is self.__class__:
            data = {
                k: v
                for k, v in super().to_representation(obj).items()
                if k not in obj.polymorphic_internal_model_fields
            }
        else:
            data = serializer().to_representation(obj)

        data[self.type_field] = obj.__class__.__name__

        return data

    def to_internal_value(self, data):
        data_type = data.get(self.type_field)

        validated_data = super().to_internal_value(data)

        validated_data[self.type_field] = data_type

        return validated_data

    @transaction.atomic
    def create(self, validated_data):
        data_type = validated_data.pop(self.type_field, None)
        serializer = self.get_serializer_from_type(data_type)(validated_data)

        if serializer.__class__ is self.__class__:
            return super().create(validated_data)
        else:
            return serializer.create(validated_data)


class FieldSerializer(ModelSerializer):
    class Meta:
        model = Field
        exclude = ("source",)
        read_only_fields = ("name", "sample", "source")


class SourceSerializer(PolymorphicModelSerializer):
    fields = FieldSerializer(many=True, required=False)
    status = SerializerMethodField()
    slug = SlugField(max_length=255, read_only=True)

    class Meta:
        fields = "__all__"
        model = Source

    def _update_fields(self, source):
        if source.run_sync_method("update_fields", success_state="NEED_SYNC").result:
            return source
        raise ValidationError("Fields update failed")

    @transaction.atomic
    def create(self, validated_data):
        # Fields can't be defined at source creation
        validated_data.pop("fields", None)
        source = super().create(validated_data)
        return self._update_fields(source)

    @transaction.atomic
    def update(self, instance, validated_data):
        validated_data.pop("fields")

        source = super().update(instance, validated_data)

        self._update_fields(source)

        for field_data in self.get_initial().get("fields", []):

            try:
                instance = source.fields.get(name=field_data.get("name"))
                serializer = FieldSerializer(instance=instance, data=field_data)
                if serializer.is_valid():
                    serializer.save()
                else:
                    raise ValidationError("Field configuration is not valid")
            except Field.DoesNotExist:
                pass

        return source

    def get_status(self, instance):
        return instance.get_status()


class PostGISSourceSerializer(SourceSerializer):
    id_field = CharField(required=False)
    geom_field = CharField(required=False, allow_null=True)

    def _get_connection(self, data):
        conn = psycopg2.connect(
            user=data.get("db_username"),
            password=data.get("db_password"),
            host=data.get("db_host"),
            port=data.get("db_port", 5432),
            dbname=data.get("db_name"),
        )
        return conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

    def _first_record(self, data):
        cursor = self._get_connection(data)
        query = "SELECT * FROM ({}) q LIMIT 1"
        cursor.execute(sql.SQL(query).format(sql.SQL(data["query"])))
        return cursor.fetchone()

    def _validate_geom(self, data):
        """ Validate that geom_field exists else try to find it in source
        """
        first_record = self._first_record(data)

        if data.get("geom_field") is None:
            for k, v in first_record.items():
                try:
                    geom = GEOSGeometry(v)
                    if geom.geom_typeid == data.get("geom_type"):
                        data["geom_field"] = k
                        break
                except Exception:
                    pass

            else:
                geomtype_name = GeometryTypes(data.get("geom_type")).name
                raise ValidationError(f"No geom field found of type {geomtype_name}")
        elif data.get("geom_field") not in first_record:
            raise ValidationError("Field does not exist in source")

        return data

    def _validate_query_connection(self, data):
        """ Check if connection informations are valid or not, trying to
        connect to the Pg server and executing the query
        """
        try:
            self._first_record(data)
        except Exception:
            raise ValidationError("Connection informations or query are not valid")

    def validate(self, data):
        self._validate_query_connection(data)
        data = self._validate_geom(data)

        return super().validate(data)

    class Meta:
        model = PostGISSource
        fields = "__all__"
        extra_kwargs = {"db_password": {"write_only": True}}


class FileSourceSerializer(SourceSerializer):
    filename = SerializerMethodField()

    def to_internal_value(self, data):
        if len(data.get("file", [])) > 0:
            data["file"] = data["file"][0]

        return super().to_internal_value(data)

    def get_filename(self, instance):
        if instance.file:
            return basename(instance.file.name)

    class Meta:
        model = None


class GeoJSONSourceSerializer(FileSourceSerializer):
    class Meta:
        model = GeoJSONSource
        fields = "__all__"
        extra_kwargs = {"file": {"write_only": True}}


class ShapefileSourceSerializer(FileSourceSerializer):
    class Meta:
        model = ShapefileSource
        fields = "__all__"
        extra_kwargs = {"file": {"write_only": True}}


class CommandSourceSerializer(SourceSerializer):
    class Meta:
        model = CommandSource
        fields = "__all__"
        extra_kwargs = {"command": {"read_only": True}}


class WMTSSourceSerialize(SourceSerializer):
    minzoom = IntegerField(min_value=0, max_value=24, allow_null=True, default=0)
    maxzoom = IntegerField(min_value=0, max_value=24, allow_null=True, default=24)
    geom_type = CharField(
        required=False, allow_null=True, default=GeometryTypes.Undefined.value
    )

    class Meta:
        model = WMTSSource
        fields = "__all__"


class CSVSourceSerializer(FileSourceSerializer):
    coordinate_reference_system = CharField(required=True)
    encoding = CharField(required=True)
    field_separator = CharField(required=True)
    decimal_separator = CharField(required=True)
    char_delimiter = CharField(required=True)
    coordinates_field = CharField(required=True)
    number_lines_to_ignore = IntegerField(required=True)

    use_header = BooleanField(required=False, default=False)
    ignore_columns = BooleanField(required=False, default=False)
    latitude_field = CharField(required=False)
    longitude_field = CharField(required=False)
    latlong_field = CharField(required=False)
    coordinates_field_count = CharField(required=False)
    coordinates_separator = CharField(required=False)
    geom_type = ChoiceField(
        default=GeometryTypes.Point.value, choices=GeometryTypes.choices()
    )

    class Meta:
        model = CSVSource
        fields = "__all__"
        extra_kwargs = {
            "file": {"write_only": True},
            "settings": {"write_only": True},
        }

    def to_internal_value(self, data):
        validated_data = super().to_internal_value(data)
        validated_data["settings"] = {
            "coordinate_reference_system": validated_data.pop(
                "coordinate_reference_system"
            ),
            "encoding": validated_data.pop("encoding"),
            "field_separator": validated_data.pop("field_separator"),
            "decimal_separator": validated_data.pop("decimal_separator"),
            "char_delimiter": validated_data.pop("char_delimiter"),
            "coordinates_field": validated_data.get("coordinates_field"),
            "number_lines_to_ignore": validated_data.pop("number_lines_to_ignore"),
            "use_header": validated_data.pop("use_header"),
            "ignore_columns": validated_data.pop("ignore_columns"),
        }
        if validated_data.get("coordinates_field") == "one_column":
            validated_data["settings"].update(
                {
                    "latlong_field": validated_data.pop("latlong_field"),
                    "coordinates_field_count": validated_data.pop(
                        "coordinates_field_count"
                    ),
                    "coordinates_separator": validated_data.pop(
                        "coordinates_separator"
                    ),
                }
            )

        elif validated_data.get("coordinates_field") == "two_columns":
            validated_data["settings"].update(
                {
                    "latitude_field": validated_data.pop("latitude_field"),
                    "longitude_field": validated_data.pop("longitude_field"),
                }
            )
        validated_data.pop("coordinates_field")
        return validated_data

    def to_representation(self, obj):
        data = super().to_representation(obj)
        if data.get("coordinates_field") == "one_column":
            data.pop("latitude_field")
            data.pop("longitude_field")

        if data.get("coordinates_field") == "two_columns":
            data.pop("latlong_field")
            data.pop("coordinates_field_count")
            data.pop("coordinates_separator")
        return data

    def validate(self, data):
        validated_data = super().validate(data)
        if data["settings"]["coordinates_field"] == "one_column":
            if not data["settings"].get("latlong_field"):
                raise ValidationError(
                    _(
                        "latlong_field must be defined when coordinates are set to one column"
                    )
                )
            if not data["settings"].get("coordinates_field_count"):
                raise ValidationError(
                    _(
                        "Coordinates order must be specified when coordinates are set to one column"
                    )
                )
            if not data["settings"].get("coordinates_separator"):
                raise ValidationError(
                    _(
                        "Coordinates separator must be specified when coordinates are set to one column"
                    )
                )
        elif data["settings"]["coordinates_field"] == "two_columns":
            if not data["settings"].get("latitude_field"):
                raise ValidationError(
                    _(
                        "Latitude field must be specified when coordinates are set to two columns"
                    )
                )
            if not data["settings"].get("longitude_field"):
                raise ValidationError(
                    _(
                        "Longitude field must be specified when coordinates are set to two columns"
                    )
                )
        else:
            raise ValidationError(_("Incorrect value for coordinates field"))
        return validated_data



import logging

from celery import shared_task, states
from celery.exceptions import Ignore
from django.apps import apps

logger = logging.getLogger(__name__)


def set_failure_state(task, method, message):

    # Failure messaging needs to be formed as expected by celery API
    logger.warning(message)
    task.update_state(
        state=states.FAILURE,
        meta={
            "method": method,
            "exc_type": Exception.__name__,
            "exc_message": [message],
        },
    )


@shared_task(bind=True)
def run_model_object_method(self, app, model, pk, method, success_state=states.SUCCESS):
    self.update_state(state=states.STARTED)

    Model = apps.get_app_config(app).get_model(model)

    try:
        obj = Model.objects.get(pk=pk)
        state = {"action": method, **getattr(obj, method)()}

        self.update_state(state=success_state, meta=state)

    except Model.DoesNotExist:
        set_failure_state(self, method, f"{Model}'s object with pk {pk} doesn't exist")

    except AttributeError as e:
        set_failure_state(self, method, f"{method} doesn't exist for object {obj}: {e}")
        logger.error(e, exc_info=True)

    except Exception as e:
        if hasattr(e, "message"):
            message = e.message
        else:
            message = f"{e}"
        set_failure_state(self, method, message)
        logger.error(e, exc_info=True)

    raise Ignore()



# Generated by Django 2.0.13 on 2019-07-05 12:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0011_auto_20190628_1449")]

    operations = [
        migrations.AddField(
            model_name="field", name="level", field=models.IntegerField(default=0)
        )
    ]



import copy
import heapq
import logging
from datetime import datetime, timedelta

from celery import schedules
from celery.beat import ScheduleEntry, Scheduler, event_t
from django.utils import timezone
from django_geosource.models import PostGISSource

logger = logging.getLogger(__name__)


class SourceEntry(ScheduleEntry):
    def __init__(self, source, *args, **kwargs):
        self.source = source
        self.name = source.slug
        super().__init__(*args, **kwargs)

    def is_due(self):
        logger.info(f"Is {self.source} due to refresh ?")

        if self.source.refresh > 0:
            next_run = self.last_run_at + timedelta(minutes=self.source.refresh)
            logger.debug(
                f"Refresh : {self.source.refresh} | Next run : {next_run} | Now : {timezone.now()}"
            )
            if next_run < timezone.now():
                logger.info("Source is due to refresh")
                return schedules.schedstate(True, 10.0)
            else:
                logger.debug("Source is NOT due to refresh, let's wait")
                return schedules.schedstate(False, 10.0)

        logger.info("The refresh is disabled for this source, check again later")
        return schedules.schedstate(False, 600.0)

    def run_task(self):
        logger.info(f"Launch refresh_data for source {self.source}")
        try:
            self.source.run_async_method("refresh_data")
        except Exception as e:
            logger.error(f"Refresh data couldn't be launched: {e}")

    def __next__(self, last_run_at=None):
        """Return new instance, with date and count fields updated."""
        return self.__class__(
            **dict(
                self,
                source=self.source,
                last_run_at=last_run_at or self.default_now(),
                total_run_count=self.total_run_count + 1,
            )
        )

    next = __next__


class GeosourceScheduler(Scheduler):
    """Database-backed Beat Scheduler."""

    Entry = SourceEntry
    _initial = True
    _last_sync = None

    TICK_DELAY = 60

    def all_entries(self):
        s = {}
        for source in PostGISSource.objects.exclude(refresh__lte=0).order_by(
            "-refresh"
        ):
            try:
                s[source.slug] = self.Entry(source, app=self.app)
            except ValueError:
                pass
        return s

    def reserve(self, entry):
        new_entry = self.schedule[entry.source.slug] = next(entry)
        return new_entry

    @property
    def schedule(self):
        if self.should_sync():
            logger.debug("Resync schedule entries")
            self.sync()
            self._schedule = self.all_entries()

        return self._schedule

    def sync(self):
        self._last_sync = timezone.now().timestamp()
        super().sync()

    def should_sync(self):
        last_update = PostGISSource.objects.order_by("-updated_at").first()
        if not self._last_sync or (
            last_update
            and (
                timezone.make_aware(
                    datetime.fromtimestamp(self._last_sync),
                    timezone.get_default_timezone(),
                )
                < last_update.updated_at
            )
        ):
            return True
        return False

    def apply_entry(self, entry, producer=None):
        logger.info(f"Scheduler: Sending due task {entry.source.name}")
        entry.run_task()

    def tick(
        self, event_t=event_t, min=min, heappop=heapq.heappop, heappush=heapq.heappush
    ):

        logger.debug("Ticking")

        max_interval = self.max_interval

        if self._heap is None or not self.schedules_equal(
            self.old_schedulers, self.schedule
        ):
            self.old_schedulers = copy.copy(self.schedule)
            self.populate_heap()

        H = self._heap

        if not H:
            logger.info("There is no source to synchronize")
            return max_interval

        event = H[0]
        time_to_run, priority, entry = event
        time_to_run = datetime.fromtimestamp(time_to_run)

        if time_to_run < datetime.now():
            is_due, next_time_to_run = self.is_due(entry)
            verify = heappop(H)

            if is_due and verify is event:
                self.apply_entry(entry, producer=self.producer)

            next_entry = self.reserve(entry)
            heappush(
                H,
                event_t(self._when(next_entry, next_time_to_run), priority, next_entry),
            )

        return self.TICK_DELAY / len(H)



from django.conf import settings

# Max time a task can be running until another one can be runned.
# This is to prevent when a task is blocked.
MAX_TASK_RUNTIME = getattr(settings, "GEOSOURCE_MAX_TASK_RUNTIME", 24)



from rest_framework import status
from rest_framework.decorators import action
from rest_framework.parsers import JSONParser
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from .models import Source
from .parsers import NestedMultipartJSONParser
from .permissions import SourcePermission
from .serializers import SourceSerializer


class SourceModelViewset(ModelViewSet):
    model = Source
    parser_classes = (JSONParser, NestedMultipartJSONParser)
    serializer_class = SourceSerializer
    permission_classes = (SourcePermission,)
    ordering_fields = (
        "name",
        "polymorphic_ctype__model",
        "geom_type",
        "id",
        "slug",
    )
    filter_fields = (
        "polymorphic_ctype",
        "geom_type",
    )
    search_fields = ["name"]

    def get_queryset(self):
        return self.model.objects.all()

    @action(detail=True, methods=["get"])
    def refresh(self, request, pk):
        """ Schedule a refresh now
        """

        source = self.get_object()

        refresh_job = source.run_async_method("refresh_data")
        if refresh_job:
            return Response(data=source.get_status(), status=status.HTTP_202_ACCEPTED)

        return Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR)



import logging

from django.contrib.auth.models import Group
from django.contrib.gis.geos import GEOSGeometry
from geostore.models import Layer, LayerGroup

logger = logging.getLogger(__name__)


def layer_callback(geosource):

    group_name = geosource.settings.pop("group", "reference")

    defaults = {
        "settings": geosource.settings,
    }

    layer, _ = Layer.objects.get_or_create(name=geosource.slug, defaults=defaults)

    layer_groups = Group.objects.filter(pk__in=geosource.settings.get("groups", []))

    if set(layer.authorized_groups.all()) != set(layer_groups):
        layer.authorized_groups.set(layer_groups)

    if not layer.layer_groups.filter(name=group_name).exists():
        group, _ = LayerGroup.objects.get_or_create(name=group_name)
        group.layers.add(layer)

    return layer


def feature_callback(geosource, layer, identifier, geometry, attributes):
    # Force converting geometry to 4326 projection
    try:
        geom = GEOSGeometry(geometry)
        geom.transform(4326)
        return layer.features.update_or_create(
            identifier=identifier, defaults={"properties": attributes, "geom": geom}
        )[0]
    except (TypeError, ValueError):
        logger.warning(
            f"One record was ignored from source, because of invalid geometry: {attributes}"
        )
        return None


def clear_features(geosource, layer, begin_date):
    return layer.features.filter(updated_at__lt=begin_date).delete()


def delete_layer(geosource):
    geosource.get_layer().features.all().delete()
    return geosource.get_layer().delete()



# Generated by Django 3.0.6 on 2020-05-06 14:44

from django.db import migrations, models
import django.db.models.deletion
import django_geosource.models


class Migration(migrations.Migration):

    dependencies = [
        ("django_geosource", "0015_auto_20190927_1445"),
    ]

    operations = [
        migrations.CreateModel(
            name="CSVSource",
            fields=[
                (
                    "source_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_geosource.Source",
                    ),
                ),
                ("file", models.FileField(upload_to="geosource/csv/%Y")),
            ],
            options={"abstract": False, "base_manager_name": "objects"},
            bases=("django_geosource.source",),
        ),
        migrations.AlterField(
            model_name="source",
            name="geom_type",
            field=models.IntegerField(
                choices=[
                    (0, django_geosource.models.GeometryTypes["Point"]),
                    (1, django_geosource.models.GeometryTypes["LineString"]),
                    (3, django_geosource.models.GeometryTypes["Polygon"]),
                    (4, django_geosource.models.GeometryTypes["MultiPoint"]),
                    (5, django_geosource.models.GeometryTypes["MultiLineString"]),
                    (6, django_geosource.models.GeometryTypes["MultiPolygon"]),
                    (7, django_geosource.models.GeometryTypes["GeometryCollection"]),
                    (8, django_geosource.models.GeometryTypes["Undefined"]),
                ]
            ),
        ),
    ]



# Generated by Django 2.0.13 on 2019-05-24 09:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="postgissource",
            name="refresh",
            field=models.IntegerField(default=-1),
        )
    ]



from django import forms
from django.core.validators import URLValidator
from django.db.models.fields import TextField
from django.utils.translation import gettext_lazy as _


class LongURLField(TextField):
    default_validators = [URLValidator()]
    description = _("URL")

    def formfield(self, **kwargs):
        # As with CharField, this will cause URL validation to be performed
        # twice.
        defaults = {"form_class": forms.URLField}
        defaults.update(kwargs)
        return super().formfield(**defaults)



from rest_framework import routers

from .views import SourceModelViewset

app_name = "geosource"

router = routers.SimpleRouter()

router.register(r"", SourceModelViewset, basename="geosource")

urlpatterns = router.urls



import json
import logging
from datetime import datetime
from enum import Enum, IntEnum, auto

import fiona
import psycopg2
import pyexcel
from celery.result import AsyncResult
from django.conf import settings
from django.contrib.gis.gdal.error import GDALException
from django.contrib.gis.geos import GEOSGeometry
from django.contrib.postgres.fields import JSONField
from django.core.management import call_command
from django.core.validators import RegexValidator, URLValidator
from django.db import models, transaction
from django.utils.text import slugify
from polymorphic.models import PolymorphicModel
from psycopg2 import sql

from .callbacks import get_attr_from_path

# from .celery import app as celery_app
from .fields import LongURLField
from .mixins import CeleryCallMethodsMixin
from .signals import refresh_data_done

logger = logging.getLogger(__name__)

# Decimal fields must be returned as float
DEC2FLOAT = psycopg2.extensions.new_type(
    psycopg2.extensions.DECIMAL.values,
    "DEC2FLOAT",
    lambda value, curs: float(value) if value is not None else None,
)
psycopg2.extensions.register_type(DEC2FLOAT)

HOST_REGEX = (
    rf"(?:{URLValidator.ipv4_re}|{URLValidator.ipv6_re}|{URLValidator.host_re})"
)


class FieldTypes(Enum):
    String = auto()
    Integer = auto()
    Float = auto()
    Boolean = auto()
    Undefined = auto()

    def _generate_next_value_(self, name, start, count, last_values):
        return name.lower()

    @classmethod
    def choices(cls):
        return [(enum.value, enum) for enum in cls]

    @classmethod
    def get_type_from_data(cls, data):
        types = {
            type(None): cls.Undefined,
            str: cls.String,
            int: cls.Integer,
            bool: cls.Boolean,
            float: cls.Float,
        }

        return types.get(type(data), cls.Undefined)


class GeometryTypes(IntEnum):
    Point = 0
    LineString = 1
    # LinearRing 2
    Polygon = 3
    MultiPoint = 4
    MultiLineString = 5
    MultiPolygon = 6
    GeometryCollection = 7
    Undefined = 8

    @classmethod
    def choices(cls):
        return [(enum.value, enum) for enum in cls]


class Source(PolymorphicModel, CeleryCallMethodsMixin):
    name = models.CharField(max_length=255, unique=True)
    slug = models.SlugField(max_length=255, unique=True)
    description = models.TextField(blank=True)

    id_field = models.CharField(max_length=255, default="id")
    geom_type = models.IntegerField(choices=GeometryTypes.choices())

    settings = JSONField(default=dict)

    task_id = models.CharField(null=True, max_length=255)
    task_date = models.DateTimeField(null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    SOURCE_GEOM_ATTRIBUTE = "_geom_"
    MAX_SAMPLE_DATA = 5

    class Meta:
        permissions = (("can_manage_sources", "Can manage sources"),)

    def get_layer(self):
        return get_attr_from_path(settings.GEOSOURCE_LAYER_CALLBACK)(self)

    def update_feature(self, *args):
        return get_attr_from_path(settings.GEOSOURCE_FEATURE_CALLBACK)(self, *args)

    def clear_features(self, layer, begin_date):
        return get_attr_from_path(settings.GEOSOURCE_CLEAN_FEATURE_CALLBACK)(
            self, layer, begin_date
        )

    def delete(self, *args, **kwargs):
        get_attr_from_path(settings.GEOSOURCE_DELETE_LAYER_CALLBACK)(self)
        return super().delete(*args, **kwargs)

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        return super().save(*args, **kwargs)

    def refresh_data(self):
        with transaction.atomic():
            layer = self.get_layer()
            begin_date = datetime.now()
            row_count = 0

            for row in self._get_records():
                geometry = row.pop(self.SOURCE_GEOM_ATTRIBUTE)
                try:
                    identifier = row[self.id_field]
                except KeyError:
                    raise Exception(
                        "Can't find identifier field in one or more records"
                    )
                self.update_feature(layer, identifier, geometry, row)
                row_count += 1

            self.clear_features(layer, begin_date)

        refresh_data_done.send_robust(
            sender=self.__class__, layer=layer.pk,
        )

        return {"count": row_count}

    @transaction.atomic
    def update_fields(self):
        records = self._get_records(50)

        fields = {}

        for record in records:
            record.pop(self.SOURCE_GEOM_ATTRIBUTE)

            for field_name, value in record.items():
                is_new = False

                if field_name not in fields:
                    field, is_new = self.fields.get_or_create(
                        name=field_name, defaults={"label": field_name}
                    )
                    field.sample = []
                    fields[field_name] = field

                if is_new or fields[field_name].data_type == FieldTypes.Undefined:
                    fields[field_name].data_type = FieldTypes.get_type_from_data(
                        value
                    ).value

                if (
                    len(fields[field_name].sample) < self.MAX_SAMPLE_DATA
                    and value is not None
                ):

                    if isinstance(value, bytes):
                        try:
                            value = value.decode()
                        except (UnicodeDecodeError, AttributeError):
                            logger.warning(
                                f"{field_name} couldn't be decoded for source {self.pk}"
                            )
                            continue

                    fields[field_name].sample.append(value)

        for field in fields.values():
            field.save()

        # Delete fields that are not anymore present
        self.fields.exclude(name__in=fields.keys()).delete()

        return {"count": len(fields)}

    def get_status(self):
        response = {}

        if self.task_id:
            task = AsyncResult(self.task_id)
            response = {"state": task.state, "done": task.date_done}

            if task.successful():
                response["result"] = task.result
            if task.failed():
                task_data = task.backend.get(task.backend.get_key_for_task(task.id))
                response.update(json.loads(task_data).get("result", {}))

        return response

    def _get_records(self, limit=None):
        raise NotImplementedError

    def __str__(self):
        return f"{self.name} - {self.__class__.__name__}"

    @property
    def type(self):
        return self.__class__


class Field(models.Model):
    source = models.ForeignKey(Source, related_name="fields", on_delete=models.CASCADE)
    name = models.CharField(max_length=255, blank=False)
    label = models.CharField(max_length=255)
    data_type = models.IntegerField(
        choices=FieldTypes.choices(), default=FieldTypes.Undefined.value
    )
    level = models.IntegerField(default=0)
    sample = JSONField(default=list)

    def __str__(self):
        return f"{self.name} ({self.source.name} - {self.data_type})"

    class Meta:
        unique_together = ["source", "name"]


class PostGISSource(Source):
    db_host = models.CharField(
        max_length=255, validators=[RegexValidator(regex=HOST_REGEX)],
    )
    db_port = models.IntegerField(default=5432)
    db_username = models.CharField(max_length=63)
    db_password = models.CharField(max_length=255)
    db_name = models.CharField(max_length=63)

    query = models.TextField()

    geom_field = models.CharField(max_length=255)

    refresh = models.IntegerField(default=-1)

    @property
    def SOURCE_GEOM_ATTRIBUTE(self):
        return self.geom_field

    @property
    def _db_connection(self):
        conn = psycopg2.connect(
            user=self.db_username,
            password=self.db_password,
            host=self.db_host,
            port=self.db_port,
            dbname=self.db_name,
        )
        return conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

    def _get_records(self, limit=None):
        cursor = self._db_connection

        query = "SELECT * FROM ({}) q "
        attrs = [sql.SQL(self.query)]
        if limit:
            query += "LIMIT {}"
            attrs.append(sql.Literal(limit))

        cursor.execute(sql.SQL(query).format(*attrs))

        return cursor


class GeoJSONSource(Source):
    file = models.FileField(upload_to="geosource/geojson/%Y/")

    def get_file_as_dict(self):
        try:
            return json.load(self.file)
        except json.JSONDecodeError:
            logger.info("Source's GeoJSON file is not valid")
            raise

    def _get_records(self, limit=None):
        geojson = self.get_file_as_dict()

        limit = limit if limit else len(geojson["features"])

        records = []
        for record in geojson["features"][:limit]:
            try:
                records.append(
                    {
                        self.SOURCE_GEOM_ATTRIBUTE: GEOSGeometry(
                            json.dumps(record["geometry"])
                        ),
                        **record["properties"],
                    }
                )
            except (ValueError, GDALException):
                raise ValueError(
                    f"One of source's record has bad geometry: {record['geometry']}"
                )

        return records


class ShapefileSource(Source):
    # Zipped ShapeFile
    file = models.FileField(upload_to="geosource/shapefile/%Y/")

    def _get_records(self, limit=None):
        with fiona.BytesCollection(self.file.read()) as shapefile:
            limit = limit if limit else len(shapefile)

            # Detect the EPSG
            _, srid = shapefile.crs.get("init", "epsg:4326").split(":")

            # Return geometries with a hack to set the correct geometry srid
            return [
                {
                    self.SOURCE_GEOM_ATTRIBUTE: GEOSGeometry(
                        GEOSGeometry(json.dumps(feature.get("geometry"))).wkt,
                        srid=int(srid),
                    ),
                    **feature.get("properties", {}),
                }
                for feature in shapefile[:limit]
            ]


class CommandSource(Source):
    command = models.CharField(max_length=255)

    @transaction.atomic
    def refresh_data(self):
        layer = self.get_layer()
        begin_date = datetime.now()
        call_command(self.command)

        self.clear_features(layer, begin_date)

        refresh_data_done.send_robust(sender=self.__class__, layer=layer.pk)

        return {"count": None}

    def _get_records(self, limit=None):
        return []


class WMTSSource(Source):
    minzoom = models.IntegerField(null=True)
    maxzoom = models.IntegerField(null=True)
    tile_size = models.IntegerField()
    url = LongURLField()

    def get_status(self):
        return {"state": "DONT_NEED"}

    def refresh_data(self):
        return {}

    def _get_records(self, limit=None):
        return []


class CSVSource(Source):
    SEPARATORS = {
        "comma": ",",
        "semicolon": ";",
        "tabulation": "\t",
        "space": " ",
        "column": ":",
        "doublequote": '"',
        "simplequote": "'",
        "point": ".",
    }
    file = models.FileField(upload_to="geosource/csv/%Y")

    def get_file_as_sheet(self):
        separator = self._get_separator(self.settings["field_separator"])
        quotechar = self._get_separator(self.settings["char_delimiter"])
        try:
            return pyexcel.get_sheet(
                file_name=self.file.path,
                delimiter=separator,
                encoding=self.settings["encoding"],
                quotechar=quotechar,
            )
        except pyexcel.exceptions.FileTypeNotSupported as err:
            msg = "Source's CSV file is not valid"
            logger.info(msg)
            err.message = msg  # new message for the user
            raise

    def _get_records(self, limit=None):
        sheet = self.get_file_as_sheet()
        if self.settings.get("use_header"):
            sheet.name_columns_by_row(0)

        ignored_columns = []
        if self.settings.get("ignore_columns"):
            ignored_columns = self._get_null_columns_indexes(sheet)

        limit = limit if limit else len(sheet)

        records = []
        srid = self._get_srid()
        for row in sheet:
            if self.settings["coordinates_field"] == "two_columns":
                lat_field = self.settings["latitude_field"]
                lng_field = self.settings["longitude_field"]

                x, y = self._extract_coordinates(
                    row, sheet.colnames, [lng_field, lat_field]
                )
                ignored_field = (row.index(x), row.index(y), *ignored_columns)
            else:
                lnglat_field = self.settings["latlong_field"]
                x, y = self._extract_coordinates(row, sheet.colnames, [lnglat_field])
                coord_fields = (
                    (sheet.colnames.index(lnglat_field),)
                    if self.settings.get("use_header")
                    else (int(lnglat_field),)
                )
                ignored_field = (*coord_fields, *ignored_columns)

            cells = self._get_cells(sheet, row, ignored_field)
            records.append(
                {
                    self.SOURCE_GEOM_ATTRIBUTE: GEOSGeometry(
                        f"Point({x} {y})", srid=srid
                    ),
                    **cells,
                }
            )
        return records

    def _extract_coordinates(self, row, colnames, fields):
        coords = []
        for field in fields:
            # if no header, we expect index for the columns has been provided
            field_index = (
                colnames.index(field) if self.settings.get("use_header") else int(field)
            )
            c = row[field_index]
            coords.append(c)
        if len(coords) == 2:
            x, y = coords
        else:
            sep = self._get_separator(self.settings["coordinates_separator"])
            is_xy = self.settings["coordinates_field_count"] == "xy"
            # some fools use a reversed cartesian coordinates system (╯°□°)╯︵ ┻━┻
            x, y = coords[0].split(sep) if is_xy else coords[0].split(sep)[::-1]

        # correct formated decimal is required for GEOSGeometry
        if not self.settings["decimal_separator"] == "point":
            delimiter = self._get_separator(self.settings["decimal_separator"])
            x = x.replace(delimiter, ".")
            y = y.replace(delimiter, ".")

        return (x, y)

    def _get_null_columns_indexes(self, sheet):
        null_columns_indexes = []
        for i, column in enumerate(sheet.column):
            non_empty_cells = [cell for cell in column if cell != ""]
            (len(non_empty_cells) == 0) and null_columns_indexes.append(i)
        return null_columns_indexes

    def _get_cells(self, sheet, row, ingored_columns):
        if not self.settings.get("use_header"):
            # records names are the column index when no header was provided
            # casting to str to avoid issue (e.i id_field)
            return {
                str(i): value for i, value in enumerate(row) if i not in ingored_columns
            }

        return {
            name: value
            for i, (name, value) in enumerate(zip(sheet.colnames, row))
            if i not in ingored_columns
        }

    def _get_separator(self, name):
        return self.SEPARATORS[name]

    def _get_srid(self):
        return int(self.settings["coordinate_reference_system"].split("_")[1])

    # properties are use by serializer for representation (reading operation)
    @property
    def coordinate_reference_system(self):
        return self.settings.get("coordinate_reference_system")

    @property
    def encoding(self):
        return self.settings.get("encoding")

    @property
    def field_separator(self):
        return self.settings.get("field_separator")

    @property
    def decimal_separator(self):
        return self.settings.get("decimal_separator")

    @property
    def char_delimiter(self):
        return self.settings.get("char_delimiter")

    @property
    def coordinates_field(self):
        return self.settings.get("coordinates_field")

    @property
    def number_lines_to_ignore(self):
        return self.settings.get("number_lines_to_ignore")

    @property
    def use_header(self):
        return self.settings.get("use_header")

    @property
    def ignore_columns(self):
        return self.settings.get("ignore_columns")

    @property
    def latitude_field(self):
        return self.settings.get("latitude_field")

    @property
    def longitude_field(self):
        return self.settings.get("longitude_field")

    @property
    def latlong_field(self):
        return self.settings.get("latlong_field")

    @property
    def coordinates_field_count(self):
        return self.settings.get("coordinates_field_count")

    @property
    def coordinates_separator(self):
        return self.settings.get("coordinates_separator")



import json

from rest_framework import parsers


class NestedMultipartJSONParser(parsers.MultiPartParser):
    """
    Parser for processing multipart with json content
    """

    def parse(self, stream, media_type=None, parser_context=None):
        result = super().parse(
            stream=stream, media_type=media_type, parser_context=parser_context
        )

        data = {}
        for key, value in result.data.items():
            try:
                data[key] = json.loads(value)
            except json.JSONDecodeError:
                data[key] = value

        return parsers.DataAndFiles(data, result.files)



from datetime import timedelta

from celery import states
from django.utils.timezone import now
from django_geosource.tasks import run_model_object_method
from rest_framework.exceptions import MethodNotAllowed

from .app_settings import MAX_TASK_RUNTIME


class CeleryCallMethodsMixin:

    DONE_STATUSES = ("SUCCESS", "FAILURE", "NEED_SYNC", None)

    def update_status(self, task):
        self.task_id = task.task_id
        self.task_date = now()
        self.save()

    @property
    def can_sync(self):
        """ Property containing a boolean that tell if the state allow to run a sync
        """
        status = self.get_status()

        return status.get("state") in self.DONE_STATUSES or (
            status.get("state") not in self.DONE_STATUSES
            and self.task_date is not None
            and self.task_date < now() - timedelta(hours=MAX_TASK_RUNTIME)
        )

    def run_async_method(self, method, success_state=states.SUCCESS, force=False):
        """ Schedule an async task that will be runned by celery.
        Raises an error if a task is already running or scheduled, can be forced with
        `force` argument.
        """
        if self.can_sync or force:
            task_job = run_model_object_method.delay(
                self._meta.app_label,
                self.__class__.__name__,
                self.pk,
                method,
                success_state,
            )

            self.update_status(task_job)
            return task_job

        raise MethodNotAllowed("One job is still running on this source")

    def run_sync_method(self, method, success_state=states.SUCCESS):
        """ Run an object method in a synchrone mode.
        The success state of the task can be defined with the `success_state` argument.
        """
        task_job = run_model_object_method.apply(
            (
                self._meta.app_label,
                self.__class__.__name__,
                self.pk,
                method,
                success_state,
            )
        )
        self.update_status(task_job)
        return task_job



# Generated by Django 2.0.13 on 2019-06-03 13:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0004_auto_20190529_1647")]

    operations = [
        migrations.AlterField(
            model_name="source",
            name="status",
            field=models.CharField(max_length=255, null=True),
        )
    ]



from django.core.management import BaseCommand
from django_geosource.models import Source


class Command(BaseCommand):
    help = "Launch resync of all available sources"

    def add_arguments(self, parser):
        parser.add_argument(
            "--force",
            dest="force",
            action="store_true",
            help="Force sync even if already launched",
        )
        parser.add_argument(
            "--ask",
            dest="ask",
            action="store_true",
            help="Ask confirmation before syncing a source",
        )
        parser.add_argument(
            "--sync", dest="sync", action="store_true", help="Run in sync"
        )

    def handle(self, *args, **options):
        for source in Source.objects.all():
            if options["ask"]:
                resp = (
                    input(
                        f"Do you like to refresh the source {source}<{source.id}>? [y|N]"
                    )
                    or "n"
                )
            else:
                resp = "y"

            if resp.lower() == "y":
                if options["sync"]:
                    print(f"Refreshing source {source}<{source.id}>...")
                    source.refresh_data()
                else:
                    print(f"Schedule refresh for source {source}<{source.id}>...")
                    source.run_async_method("refresh_data", force=options["force"])



# Generated by Django 2.0.13 on 2019-05-28 14:29

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0002_auto_20190524_1137")]

    operations = [
        migrations.AlterModelOptions(
            name="source",
            options={"permissions": (("can_manage_sources", "Can manage sources"),)},
        )
    ]



from rest_framework import permissions


class SourcePermission(permissions.BasePermission):
    def has_permission(self, request, view):
        return request.user.has_perm("django_geosource.can_manage_sources")



# Generated by Django 2.0.13 on 2019-06-28 12:49

import django_geosource.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0010_wmtssource")]

    operations = [
        migrations.AlterField(
            model_name="wmtssource",
            name="url",
            field=django_geosource.fields.LongURLField(),
        )
    ]



# Generated by Django 2.1.9 on 2019-06-27 08:20

from django.db import migrations, models
from django.utils.text import slugify


def set_default(apps, schema_editor):
    Source = apps.get_model("django_geosource", "Source")
    for source in Source.objects.all():
        source.slug = slugify(source.name)
        source.save()


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0008_source_settings")]

    operations = [
        migrations.AddField(
            model_name="source",
            name="slug",
            field=models.SlugField(null=True, max_length=255, unique=True),
            preserve_default=False,
        ),
        migrations.RunPython(set_default),
        migrations.AlterField(
            model_name="source",
            name="slug",
            field=models.SlugField(max_length=255, unique=True),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.0.13 on 2019-07-09 15:35

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0012_field_level")]

    operations = [
        migrations.CreateModel(
            name="ShapefileSource",
            fields=[
                (
                    "source_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_geosource.Source",
                    ),
                ),
                ("file", models.FileField(upload_to="geosource/shapefile/%Y/")),
            ],
            options={"abstract": False, "base_manager_name": "objects"},
            bases=("django_geosource.source",),
        ),
        migrations.AlterField(
            model_name="geojsonsource",
            name="file",
            field=models.FileField(upload_to="geosource/geojson/%Y/"),
        ),
    ]



from django.core.management import BaseCommand
from django_geosource.models import Source


class Command(BaseCommand):
    help = "Launch resync of all available sources"

    def add_arguments(self, parser):
        parser.add_argument(
            "-pk", type=int, action="store", help="Pk of the source to sync"
        )
        parser.add_argument(
            "--sync", dest="sync", action="store_true", help="Run in sync"
        )

    def handle(self, *args, **options):
        source = Source.objects.get(id=options["pk"])
        if options["sync"]:
            print(f"Refreshing source {source}<{source.id}>...")
            source.refresh_data()
        else:
            print(f"Schedule refresh for source {source}<{source.id}>...")
            source.run_async_method("refresh_data")



# Generated by Django 2.0.13 on 2019-08-21 14:48

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0013_auto_20190709_1735")]

    operations = [
        migrations.AddField(
            model_name="source",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="source",
            name="updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
    ]



# Generated by Django 2.0.13 on 2019-05-29 14:47

import django_geosource.models
from django.db import migrations, models
from django_geosource.models import FieldTypes


def fix_data_types(apps, schema_editor):
    Field = apps.get_model("django_geosource", "Field")
    for field in Field.objects.all():
        field.data_type = getattr(FieldTypes, "String").value
        field.save()


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0003_auto_20190528_1629")]

    operations = [
        migrations.RunPython(fix_data_types),
        migrations.AlterField(
            model_name="field",
            name="data_type",
            field=models.IntegerField(
                choices=[
                    (1, django_geosource.models.FieldTypes(1)),
                    (2, django_geosource.models.FieldTypes(2)),
                    (3, django_geosource.models.FieldTypes(3)),
                    (4, django_geosource.models.FieldTypes(4)),
                    (5, django_geosource.models.FieldTypes(5)),
                ],
                default=5,
            ),
        ),
    ]



# Generated by Django 2.1.12 on 2019-09-27 12:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0014_auto_20190821_1648")]

    operations = [
        migrations.RenameField(
            model_name="source", old_name="status", new_name="task_id"
        ),
        migrations.AddField(
            model_name="source", name="task_date", field=models.DateTimeField(null=True)
        ),
    ]



# Generated by Django 2.0.13 on 2019-06-21 13:59

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("django_geosource", "0007_commandsource")]

    operations = [
        migrations.AddField(
            model_name="source",
            name="settings",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        )
    ]



# Generated by Django 2.0.13 on 2019-05-22 15:17

import django.contrib.postgres.fields.jsonb
import django.core.validators
import django.db.models.deletion
import django_geosource.mixins
import django_geosource.models
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [("contenttypes", "0002_remove_content_type_name")]

    operations = [
        migrations.CreateModel(
            name="Field",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("label", models.CharField(max_length=255)),
                (
                    "data_type",
                    models.CharField(
                        choices=[
                            (1, django_geosource.models.FieldTypes(1)),
                            (2, django_geosource.models.FieldTypes(2)),
                            (3, django_geosource.models.FieldTypes(3)),
                            (4, django_geosource.models.FieldTypes(4)),
                            (5, django_geosource.models.FieldTypes(5)),
                        ],
                        max_length=255,
                    ),
                ),
                ("sample", django.contrib.postgres.fields.jsonb.JSONField(default=[])),
            ],
        ),
        migrations.CreateModel(
            name="Source",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
                ("description", models.TextField(blank=True)),
                ("id_field", models.CharField(default="id", max_length=255)),
                (
                    "geom_type",
                    models.IntegerField(
                        choices=[
                            (0, django_geosource.models.GeometryTypes(0)),
                            (1, django_geosource.models.GeometryTypes(1)),
                            (3, django_geosource.models.GeometryTypes(3)),
                            (4, django_geosource.models.GeometryTypes(4)),
                            (5, django_geosource.models.GeometryTypes(5)),
                            (6, django_geosource.models.GeometryTypes(6)),
                            (7, django_geosource.models.GeometryTypes(7)),
                        ]
                    ),
                ),
                ("status", models.NullBooleanField(default=None)),
            ],
            options={"abstract": False, "base_manager_name": "objects"},
            bases=(models.Model, django_geosource.mixins.CeleryCallMethodsMixin),
        ),
        migrations.CreateModel(
            name="GeoJSONSource",
            fields=[
                (
                    "source_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_geosource.Source",
                    ),
                ),
                ("file", models.FileField(upload_to="geosource/")),
            ],
            options={"abstract": False, "base_manager_name": "objects"},
            bases=("django_geosource.source",),
        ),
        migrations.CreateModel(
            name="PostGISSource",
            fields=[
                (
                    "source_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_geosource.Source",
                    ),
                ),
                (
                    "db_host",
                    models.CharField(
                        max_length=255,
                        validators=[
                            django.core.validators.RegexValidator(
                                regex="(?:(?:25[0-5]|2[0-4]\\d|[0-1]?\\d?\\d)(?:\\.(?:25[0-5]|2[0-4]\\d|[0-1]?\\d?\\d)){3}|\\[[0-9a-f:\\.]+\\]|([a-z¡-\uffff0-9](?:[a-z¡-\uffff0-9-]{0,61}[a-z¡-\uffff0-9])?(?:\\.(?!-)[a-z¡-\uffff0-9-]{1,63}(?<!-))*\\.(?!-)(?:[a-z¡-\uffff-]{2,63}|xn--[a-z0-9]{1,59})(?<!-)\\.?|localhost))"
                            )
                        ],
                    ),
                ),
                ("db_port", models.IntegerField(default=5432)),
                ("db_username", models.CharField(max_length=63)),
                ("db_password", models.CharField(max_length=255)),
                ("db_name", models.CharField(max_length=63)),
                ("query", models.TextField()),
                ("geom_field", models.CharField(max_length=255)),
                ("refresh", models.IntegerField()),
            ],
            options={"abstract": False, "base_manager_name": "objects"},
            bases=("django_geosource.source",),
        ),
        migrations.AddField(
            model_name="source",
            name="polymorphic_ctype",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="polymorphic_django_geosource.source_set+",
                to="contenttypes.ContentType",
            ),
        ),
        migrations.AddField(
            model_name="field",
            name="source",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="fields",
                to="django_geosource.Source",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="field", unique_together={("source", "name")}
        ),
    ]
